@@ -198,7 +198,7 @@
 //! designed to be as cheap as possible.
 //!
 //! With an [`Instance`] you can then invoke various exports and interact with
-//! the WebAssembly module. Once the request is finished the [`Store`],
+//! the WebAssembly module. Once the request is finished, the [`Store`]
 //! is dropped and everything will be deallocated. Note that if the same
 //! [`Store`] were used for every request then that would have all requests
 //! sharing resources and nothing would ever get deallocated, causing memory
@@ -259,7 +259,7 @@
 //!   asynchronously.
 //!
 //! * `jitdump` - Enabled by default, this feature compiles in support for the
-//!   jitdump runtime profilng format. The profiler can be selected with
+//!   jitdump runtime profiling format. The profiler can be selected with
 //!   [`Config::profiler`].
 //!
 //! * `vtune` - Not enabled by default, this feature compiles in support for
@@ -440,7 +440,7 @@ fn _assert_send_sync() {
 
     #[cfg(feature = "async")]
     fn _call_async(s: &mut Store<()>, f: Func) {
-        _assert_send(f.call_async(&mut *s, &[]))
+        _assert_send(f.call_async(&mut *s, &[], &mut []))
     }
     #[cfg(feature = "async")]
     fn _typed_call_async(s: &mut Store<()>, f: TypedFunc<(), ()>) {
