@@ -88,6 +88,7 @@ use std::mem::{self, ManuallyDrop};
 use std::ops::{Deref, DerefMut};
 use std::pin::Pin;
 use std::ptr;
+use std::sync::atomic::AtomicU64;
 use std::sync::Arc;
 use std::task::{Context, Poll};
 use wasmtime_runtime::{
@@ -258,7 +259,12 @@ pub struct StoreOpaque {
     signal_handler: Option<Box<SignalHandler<'static>>>,
     externref_activations_table: VMExternRefActivationsTable,
     modules: ModuleRegistry,
+
+    // See documentation on `StoreOpaque::lookup_trampoline` for what these
+    // fields are doing.
     host_trampolines: HashMap<VMSharedSignatureIndex, VMTrampoline>,
+    host_func_trampolines_registered: usize,
+
     // Numbers of resources instantiated in this store, and their limits
     instance_count: usize,
     instance_limit: usize,
@@ -272,6 +278,7 @@ pub struct StoreOpaque {
     #[cfg(feature = "async")]
     async_state: AsyncState,
     out_of_gas_behavior: OutOfGas,
+    epoch_deadline_behavior: EpochDeadline,
     store_data: StoreData,
     default_callee: InstanceHandle,
 
@@ -379,6 +386,18 @@ enum OutOfGas {
     },
 }
 
+/// What to do when the engine epoch reaches the deadline for a Store
+/// during execution of a function using that store.
+#[derive(Copy, Clone)]
+enum EpochDeadline {
+    /// Return early with a trap.
+    Trap,
+    /// Extend the deadline by the specified number of ticks after
+    /// yielding to the async executor loop.
+    #[cfg(feature = "async")]
+    YieldAndExtendDeadline { delta: u64 },
+}
+
 impl<T> Store<T> {
     /// Creates a new [`Store`] to be associated with the given [`Engine`] and
     /// `data` provided.
@@ -399,6 +418,7 @@ impl<T> Store<T> {
         // part of `Func::call` to guarantee that the `callee: *mut VMContext`
         // is never null.
         let default_callee = unsafe {
+            let module = Arc::new(wasmtime_environ::Module::default());
             OnDemandInstanceAllocator::default()
                 .allocate(InstanceAllocationRequest {
                     host_state: Box::new(()),
@@ -406,12 +426,15 @@ impl<T> Store<T> {
                     functions,
                     shared_signatures: None.into(),
                     imports: Default::default(),
-                    module: Arc::new(wasmtime_environ::Module::default()),
+                    module: &module,
+                    unique_id: None,
+                    memfds: None,
                     store: StorePtr::empty(),
                     wasm_data: &[],
                 })
                 .expect("failed to allocate default callee")
         };
+
         let mut inner = Box::new(StoreInner {
             inner: StoreOpaque {
                 _marker: marker::PhantomPinned,
@@ -422,6 +445,7 @@ impl<T> Store<T> {
                 externref_activations_table: VMExternRefActivationsTable::new(),
                 modules: ModuleRegistry::default(),
                 host_trampolines: HashMap::default(),
+                host_func_trampolines_registered: 0,
                 instance_count: 0,
                 instance_limit: crate::DEFAULT_INSTANCE_LIMIT,
                 memory_count: 0,
@@ -435,6 +459,7 @@ impl<T> Store<T> {
                     current_poll_cx: UnsafeCell::new(ptr::null_mut()),
                 },
                 out_of_gas_behavior: OutOfGas::Trap,
+                epoch_deadline_behavior: EpochDeadline::Trap,
                 store_data: StoreData::new(),
                 default_callee,
                 hostcall_val_storage: Vec::new(),
@@ -809,6 +834,86 @@ impl<T> Store<T> {
         self.inner
             .out_of_fuel_async_yield(injection_count, fuel_to_inject)
     }
+
+    /// Sets the epoch deadline to a certain number of ticks in the future.
+    ///
+    /// When the Wasm guest code is compiled with epoch-interruption
+    /// instrumentation
+    /// ([`Config::epoch_interruption()`](crate::Config::epoch_interruption)),
+    /// and when the `Engine`'s epoch is incremented
+    /// ([`Engine::increment_epoch()`](crate::Engine::increment_epoch))
+    /// past a deadline, execution can be configured to either trap or
+    /// yield and then continue.
+    ///
+    /// This deadline is always set relative to the current epoch:
+    /// `delta_beyond_current` ticks in the future. The deadline can
+    /// be set explicitly via this method, or refilled automatically
+    /// on a yield if configured via
+    /// [`epoch_deadline_async_yield_and_update()`](Store::epoch_deadline_async_yield_and_update). After
+    /// this method is invoked, the deadline is reached when
+    /// [`Engine::increment_epoch()`] has been invoked at least
+    /// `ticks_beyond_current` times.
+    ///
+    /// See documentation on
+    /// [`Config::epoch_interruption()`](crate::Config::epoch_interruption)
+    /// for an introduction to epoch-based interruption.
+    pub fn set_epoch_deadline(&mut self, ticks_beyond_current: u64) {
+        self.inner.set_epoch_deadline(ticks_beyond_current);
+    }
+
+    /// Configures epoch-deadline expiration to trap.
+    ///
+    /// When epoch-interruption-instrumented code is executed on this
+    /// store and the epoch deadline is reached before completion,
+    /// with the store configured in this way, execution will
+    /// terminate with a trap as soon as an epoch check in the
+    /// instrumented code is reached.
+    ///
+    /// This behavior is the default if the store is not otherwise
+    /// configured via
+    /// [`epoch_deadline_trap()`](Store::epoch_deadline_trap) or
+    /// [`epoch_deadline_async_yield_and_update()`](Store::epoch_deadline_async_yield_and_update).
+    ///
+    /// This setting is intended to allow for coarse-grained
+    /// interruption, but not a deterministic deadline of a fixed,
+    /// finite interval. For deterministic interruption, see the
+    /// "fuel" mechanism instead.
+    ///
+    /// See documentation on
+    /// [`Config::epoch_interruption()`](crate::Config::epoch_interruption)
+    /// for an introduction to epoch-based interruption.
+    pub fn epoch_deadline_trap(&mut self) {
+        self.inner.epoch_deadline_trap();
+    }
+
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
+    /// Configures epoch-deadline expiration to yield to the async
+    /// caller and the update the deadline.
+    ///
+    /// When epoch-interruption-instrumented code is executed on this
+    /// store and the epoch deadline is reached before completion,
+    /// with the store configured in this way, execution will yield
+    /// (the future will return `Pending` but re-awake itself for
+    /// later execution) and, upon resuming, the store will be
+    /// configured with an epoch deadline equal to the current epoch
+    /// plus `delta` ticks.
+    ///
+    /// This setting is intended to allow for cooperative timeslicing
+    /// of multiple CPU-bound Wasm guests in different stores, all
+    /// executing under the control of an async executor. To drive
+    /// this, stores should be configured to "yield and update"
+    /// automatically with this function, and some external driver (a
+    /// thread that wakes up periodically, or a timer
+    /// signal/interrupt) should call
+    /// [`Engine::increment_epoch()`](crate::Engine::increment_epoch).
+    ///
+    /// See documentation on
+    /// [`Config::epoch_interruption()`](crate::Config::epoch_interruption)
+    /// for an introduction to epoch-based interruption.
+    #[cfg(feature = "async")]
+    pub fn epoch_deadline_async_yield_and_update(&mut self, delta: u64) {
+        self.inner.epoch_deadline_async_yield_and_update(delta);
+    }
 }
 
 impl<'a, T> StoreContext<'a, T> {
@@ -913,6 +1018,31 @@ impl<'a, T> StoreContextMut<'a, T> {
         self.0
             .out_of_fuel_async_yield(injection_count, fuel_to_inject)
     }
+
+    /// Sets the epoch deadline to a certain number of ticks in the future.
+    ///
+    /// For more information see [`Store::set_epoch_deadline`].
+    pub fn set_epoch_deadline(&mut self, ticks_beyond_current: u64) {
+        self.0.set_epoch_deadline(ticks_beyond_current);
+    }
+
+    /// Configures epoch-deadline expiration to trap.
+    ///
+    /// For more information see [`Store::epoch_deadline_trap`].
+    pub fn epoch_deadline_trap(&mut self) {
+        self.0.epoch_deadline_trap();
+    }
+
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
+    /// Configures epoch-deadline expiration to yield to the async
+    /// caller and the update the deadline.
+    ///
+    /// For more information see
+    /// [`Store::epoch_deadline_async_yield_and_update`].
+    #[cfg(feature = "async")]
+    pub fn epoch_deadline_async_yield_and_update(&mut self, delta: u64) {
+        self.0.epoch_deadline_async_yield_and_update(delta);
+    }
 }
 
 impl<T> StoreInner<T> {
@@ -986,14 +1116,6 @@ impl StoreOpaque {
         &mut self.store_data
     }
 
-    pub fn register_host_trampoline(
-        &mut self,
-        idx: VMSharedSignatureIndex,
-        trampoline: VMTrampoline,
-    ) {
-        self.host_trampolines.insert(idx, trampoline);
-    }
-
     pub fn interrupt_handle(&self) -> Result<InterruptHandle> {
         if self.engine.config().tunables.interruptable {
             Ok(InterruptHandle {
@@ -1046,17 +1168,89 @@ impl StoreOpaque {
         unsafe { wasmtime_runtime::gc(&self.modules, &mut self.externref_activations_table) }
     }
 
-    pub fn lookup_trampoline(&self, anyfunc: &VMCallerCheckedAnyfunc) -> VMTrampoline {
-        // Look up the trampoline with the store's trampolines (from `Func`).
-        if let Some(trampoline) = self.host_trampolines.get(&anyfunc.type_index) {
-            return *trampoline;
-        }
-
-        // Look up the trampoline with the registered modules
+    /// Looks up the corresponding `VMTrampoline` which can be used to enter
+    /// wasm given an anyfunc function pointer.
+    ///
+    /// This is a somewhat complicated implementation at this time, unfortnately.
+    /// Trampolines are a sort of side-channel of information which is
+    /// specifically juggled by the `wasmtime` crate in a careful fashion. The
+    /// sources for trampolines are:
+    ///
+    /// * Compiled modules - each compiled module has a trampoline for all
+    ///   signatures of functions that escape the module (e.g. exports and
+    ///   `ref.func`-able functions)
+    /// * `Func::new` - host-defined functions with a dynamic signature get an
+    ///   on-the-fly-compiled trampoline (e.g. JIT-compiled as part of the
+    ///   `Func::new` call).
+    /// * `Func::wrap` - host-defined functions where the trampoline is
+    ///   monomorphized in Rust and compiled by LLVM.
+    ///
+    /// The purpose of this function is that given some wasm function pointer we
+    /// need to find the trampoline for it. For compiled wasm modules this is
+    /// pretty easy, the code pointer of the function pointer will point us
+    /// at a wasm module which has a table of trampolines-by-type that we can
+    /// lookup.
+    ///
+    /// If this lookup fails, however, then we're trying to get the trampoline
+    /// for a wasm function pointer defined by the host. The trampoline isn't
+    /// actually stored in the wasm function pointer itself so we need
+    /// side-channels of information. To achieve this a lazy scheme is
+    /// implemented here based on the assumption that most trampoline lookups
+    /// happen for wasm-defined functions, not host-defined functions.
+    ///
+    /// The `Store` already has a list of all functions in
+    /// `self.store_data().funcs`, it's just not indexed in a nice fashion by
+    /// type index or similar. To solve this there's an internal map in each
+    /// store, `host_trampolines`, which maps from a type index to the
+    /// store-owned trampoline. The actual population of this map, however, is
+    /// deferred to this function itself.
+    ///
+    /// Most of the time we are looking up a Wasm function's trampoline when
+    /// calling this function, and we don't want to make insertion of a host
+    /// function into the store more expensive than it has to be. We could
+    /// update the `host_trampolines` whenever a host function is inserted into
+    /// the store, but this is a relatively expensive hash map insertion.
+    /// Instead the work is deferred until we actually look up that trampoline
+    /// in this method.
+    ///
+    /// This all means that if the lookup of the trampoline fails within
+    /// `self.host_trampolines` we lazily populate `self.host_trampolines` by
+    /// iterating over `self.store_data().funcs`, inserting trampolines as we
+    /// go. If we find the right trampoline then it's returned.
+    pub fn lookup_trampoline(&mut self, anyfunc: &VMCallerCheckedAnyfunc) -> VMTrampoline {
+        // First try to see if the `anyfunc` belongs to any module. Each module
+        // has its own map of trampolines-per-type-index and the code pointer in
+        // the `anyfunc` will enable us to quickly find a module.
         if let Some(trampoline) = self.modules.lookup_trampoline(anyfunc) {
             return trampoline;
         }
 
+        // Next consult the list of store-local host trampolines. This is
+        // primarily populated by functions created by `Func::new` or similar
+        // creation functions, host-defined functions.
+        if let Some(trampoline) = self.host_trampolines.get(&anyfunc.type_index) {
+            return *trampoline;
+        }
+
+        // If no trampoline was found then it means that it hasn't been loaded
+        // into `host_trampolines` yet. Skip over all the ones we've looked at
+        // so far and start inserting into `self.host_trampolines`, returning
+        // the actual trampoline once found.
+        for f in self
+            .store_data
+            .funcs()
+            .skip(self.host_func_trampolines_registered)
+        {
+            self.host_func_trampolines_registered += 1;
+            self.host_trampolines.insert(f.sig_index(), f.trampoline());
+            if f.sig_index() == anyfunc.type_index {
+                return f.trampoline();
+            }
+        }
+
+        // If reached this is a bug in Wasmtime. Lookup of a trampoline should
+        // only happen for wasm functions or host functions, all of which should
+        // be indexed by the above.
         panic!("trampoline missing")
     }
 
@@ -1093,13 +1287,12 @@ impl StoreOpaque {
         };
     }
 
-    /// Yields execution to the caller on out-of-gas
+    /// Yields execution to the caller on out-of-gas or epoch interruption.
     ///
     /// This only works on async futures and stores, and assumes that we're
-    /// executing on a fiber. This will yield execution back to the caller once
-    /// and when we come back we'll continue with `fuel_to_inject` more fuel.
+    /// executing on a fiber. This will yield execution back to the caller once.
     #[cfg(feature = "async")]
-    fn out_of_gas_yield(&mut self, fuel_to_inject: u64) -> Result<(), Trap> {
+    fn async_yield_impl(&mut self) -> Result<(), Trap> {
         // Small future that yields once and then returns ()
         #[derive(Default)]
         struct Yield {
@@ -1124,19 +1317,15 @@ impl StoreOpaque {
         }
 
         let mut future = Yield::default();
-        let result = unsafe { self.async_cx().block_on(Pin::new_unchecked(&mut future)) };
-        match result {
-            // If this finished successfully then we were resumed normally via a
-            // `poll`, so inject some more fuel and keep going.
-            Ok(()) => {
-                self.add_fuel(fuel_to_inject).unwrap();
-                Ok(())
-            }
-            // If the future was dropped while we were yielded, then we need to
-            // clean up this fiber. Do so by raising a trap which will abort all
-            // wasm and get caught on the other side to clean things up.
-            Err(trap) => Err(trap),
-        }
+
+        // When control returns, we have a `Result<(), Trap>` passed
+        // in from the host fiber. If this finished successfully then
+        // we were resumed normally via a `poll`, so keep going.  If
+        // the future was dropped while we were yielded, then we need
+        // to clean up this fiber. Do so by raising a trap which will
+        // abort all wasm and get caught on the other side to clean
+        // things up.
+        unsafe { self.async_cx().block_on(Pin::new_unchecked(&mut future)) }
     }
 
     fn add_fuel(&mut self, fuel: u64) -> Result<()> {
@@ -1187,6 +1376,22 @@ impl StoreOpaque {
         }
     }
 
+    fn epoch_deadline_trap(&mut self) {
+        self.epoch_deadline_behavior = EpochDeadline::Trap;
+    }
+
+    fn epoch_deadline_async_yield_and_update(&mut self, delta: u64) {
+        assert!(
+            self.async_support(),
+            "cannot use `epoch_deadline_async_yield_and_update` without enabling async support in the config"
+        );
+        #[cfg(feature = "async")]
+        {
+            self.epoch_deadline_behavior = EpochDeadline::YieldAndExtendDeadline { delta };
+        }
+        drop(delta); // suppress warning in non-async build
+    }
+
     #[inline]
     pub fn signal_handler(&self) -> Option<*const SignalHandler<'static>> {
         let handler = self.signal_handler.as_ref()?;
@@ -1535,6 +1740,10 @@ unsafe impl<T> wasmtime_runtime::Store for StoreInner<T> {
         <StoreOpaque>::vminterrupts(self)
     }
 
+    fn epoch_ptr(&self) -> *const AtomicU64 {
+        self.engine.epoch_counter() as *const _
+    }
+
     fn externref_activations_table(
         &mut self,
     ) -> (
@@ -1651,7 +1860,10 @@ unsafe impl<T> wasmtime_runtime::Store for StoreInner<T> {
                 }
                 *injection_count -= 1;
                 let fuel = *fuel_to_inject;
-                self.out_of_gas_yield(fuel)?;
+                self.async_yield_impl()?;
+                if fuel > 0 {
+                    self.add_fuel(fuel).unwrap();
+                }
                 Ok(())
             }
             #[cfg(not(feature = "async"))]
@@ -1669,6 +1881,59 @@ unsafe impl<T> wasmtime_runtime::Store for StoreInner<T> {
 
         impl std::error::Error for OutOfGasError {}
     }
+
+    fn new_epoch(&mut self) -> Result<u64, anyhow::Error> {
+        return match &self.epoch_deadline_behavior {
+            &EpochDeadline::Trap => Err(anyhow::Error::new(EpochDeadlineError)),
+            #[cfg(feature = "async")]
+            &EpochDeadline::YieldAndExtendDeadline { delta } => {
+                // Do the async yield. May return a trap if future was
+                // canceled while we're yielded.
+                self.async_yield_impl()?;
+                // Set a new deadline.
+                self.set_epoch_deadline(delta);
+
+                // Return the new epoch deadline so the Wasm code
+                // doesn't have to reload it.
+                Ok(self.get_epoch_deadline())
+            }
+        };
+
+        #[derive(Debug)]
+        struct EpochDeadlineError;
+
+        impl fmt::Display for EpochDeadlineError {
+            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+                f.write_str("epoch deadline reached during execution")
+            }
+        }
+
+        impl std::error::Error for EpochDeadlineError {}
+    }
+}
+
+impl<T> StoreInner<T> {
+    pub(crate) fn set_epoch_deadline(&mut self, delta: u64) {
+        // Set a new deadline based on the "epoch deadline delta".
+        //
+        // Safety: this is safe because the epoch deadline in the
+        // `VMInterrupts` is accessed only here and by Wasm guest code
+        // running in this store, and we have a `&mut self` here.
+        //
+        // Also, note that when this update is performed while Wasm is
+        // on the stack, the Wasm will reload the new value once we
+        // return into it.
+        let epoch_deadline = unsafe { (*self.vminterrupts()).epoch_deadline.get_mut() };
+        *epoch_deadline = self.engine().current_epoch() + delta;
+    }
+
+    fn get_epoch_deadline(&self) -> u64 {
+        // Safety: this is safe because, as above, it is only invoked
+        // from within `new_epoch` which is called from guest Wasm
+        // code, which will have an exclusive borrow on the Store.
+        let epoch_deadline = unsafe { (*self.vminterrupts()).epoch_deadline.get_mut() };
+        *epoch_deadline
+    }
 }
 
 impl<T: Default> Default for Store<T> {
